package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.taskManager.tasks.ITask
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import java.util.*


interface IWorker {
    /**
     * Adds a new task to the worker's internal tasks channel.
     */
    suspend fun addTask(task: ITask)

    // TODO: check if we can use autoclosable interface here.
    /**
     * Function used to close the underlying resources.
     */
    fun close()
}


fun getCustomThreadpoolWorker(): Worker {
    val numCores = Runtime.getRuntime().availableProcessors()
    // use a fixed size thread pool context with numCores threads
    // TODO: Threadpool needs to be closed
    @OptIn(DelicateCoroutinesApi::class)
    val threadPoolContext = newFixedThreadPoolContext(numCores, "CustomThreadWorkerPool")
    val coroutineScope = CoroutineScope(threadPoolContext)
    return Worker("Custom-ThreadPool-Worker", coroutineScope)
}

/**
 * Abstract worker class used as a base for multithreaded worker.
 * It uses a fixed thread pool as a base for the coroutineScope.
 * The size of the thread pool is equal to the number of cores.
 */
class Worker(
    private val name: String,
    private val coroutineScope: CoroutineScope = CoroutineScope(Dispatchers.Default),
    numberOfWorkers: Int = Runtime.getRuntime().availableProcessors()
) : IWorker {

    // TODO: this could be dangerous to give it unlimited memory
    private val tasksChannel = Channel<ITask>(Channel.UNLIMITED)

    // TODO: the task map should be cleared to reduce memory usage.
    //  This needs to be triggered from outside a worker as soon as
    //  all tasks related to the specific tasks are completed.
    private val taskMap = HashMap<UUID, Pair<ITask, Job>>()
    private val logger = getLogger(javaClass)

    override suspend fun addTask(task: ITask) {
        tasksChannel.send(task)
    }

    init {
        logger.info("Starting $numberOfWorkers workers.")
        repeat(numberOfWorkers) {
            logger.info("[${Thread.currentThread().name}] launching coroutine in $name")
            launchWorker(it)
        }
    }


    /**
     * Wrapper function to execute tasks provided through the channel.
     * We use a dedicated number of coroutines to watch the tasks channel.
     * Each coroutine starts a new coroutine to execute the task. This is
     * done so that the outer coroutines are not blocked by the potentially
     * very long-running inner task performed by executeTask.
     */
    private fun launchWorker(id: Int) = coroutineScope.launch {
        for (task in tasksChannel) {
            val job = coroutineScope.launch {
                logger.debug("[{}] Processor #{}-{} received {}", Thread.currentThread().name, id, name, task)
                // TODO: we need to check if we want to wrap this executeTask in a try-catch block
                try {
                    task.run()
                } catch (e: Throwable) {
                    logger.error("Exception during task execution occurred. TaskID: ${task.taskID}")
                } finally {
                    logger.debug("[{}] Processor #{}-{} finished {}", Thread.currentThread().name, id, name, task)
                    taskMap.remove(task.taskID)
                }

            }
            taskMap[task.taskID] = Pair(task, job)
        }
    }

    override fun close() {
        coroutineScope.launch {
            tasksChannel.close()
            for (job in coroutineContext[Job]!!.children) {
                job.join()
            }
            logger.info("Worker $name has been closed")
        }
    }
}
