package de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation

import de.fraunhofer.iem.dataprovider.kpi.dto.KPICreateDto
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RawValueKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryDetailsEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import java.util.*

class MetricsTask(
    val repoId: UUID,
    private val repositoryService: RepositoryService,
    private val kpiService: KPIService,
    override val responseChannel: suspend (task: Event) -> Unit
) : Task() {
    override suspend fun execute() {
        logger.info("Starting metrics task for repoId $repoId")
        val repository = repositoryService.findRepoByID(repoId)
        val repositoryDetails = repositoryService.getLatestRepositoryDetailsByRepositoryId(repoId)
        if (repositoryDetails != null && repository != null) {
            val rootKPI = generateKPITree(repositoryDetails)
            kpiService.calculateKPIs(rootKPI)
            kpiService.storeAndPurgeOld(repository, rootKPI)
        }
    }

    private fun generateKPITree(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        val rootKPI = KPITreeFactory(repositoryDetailsEntity)

        return rootKPI
    }

    private fun KPITreeFactory(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        // lowest level leaves
        val numberOfCommitsKPI = KPICreateDto(
            "Number of Commits",
            "Total number of commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfCommits!!)
        )
        val numberOfSignedCommitsKPI = KPICreateDto(
            "Number of Signed Commits",
            "Total number of signed and verified commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfSignedCommits!!)
        )
        val isDefaultBranchProtectedKPI = KPICreateDto(
            "Default Branch Protection",
            "Used to assess compliance with a standard development process. For this purpose, it is examined whether the standard development branch is protected against unintentional changes.",
            false,
            RawValueKPICalculationStrategy(if (repositoryDetailsEntity.isDefaultBranchProtected == true) 100 else 0)
        )

        val signedCommitsRatioKPI =
            KPICreateDto(
                "Commit Signature Ratio",
                "Used to assess compliance with a common and transparent development process. It is desirable that all commits are signed by their authors. Therefore, the ratio of signed commits to all commits is determined to calculate this metric.",
                false,
                RatioKPICalculationStrategy()
            )
        signedCommitsRatioKPI.addChildKPI(numberOfCommitsKPI, 1.0)
        signedCommitsRatioKPI.addChildKPI(numberOfSignedCommitsKPI, 1.0)

        // second level
        // TODO: @Jan-Niclas Please add the dependency kpis to the security KPI
        val securityKPI = KPICreateDto(
            "Security Score",
            "Assesses the security of the software provided. For this purpose, various security-relevant analyzes are carried out, which, among other things, check the external dependencies or the code for vulnerabilities.",
            false,
            AggregationKPICalculationStrategy()
        )

        val processComplianceKPI = KPICreateDto(
            "Process Compliance Score",
            "Assesses the development process of the software provided. For this purpose, the development process traceable in the repository is compared with common development standards to enable an assessment.",
            false,
            AggregationKPICalculationStrategy()
        )

        val processTransparencyKPI = KPICreateDto(
            "Process Transparency Score",
            "Assesses the transparency resp. traceability of the development process of the provided software for external parties. For this purpose, various analyzes are performed that assess the availability of information about the software development process within the repository.",
            false,
            AggregationKPICalculationStrategy()
        )
        processComplianceKPI.addChildKPI(isDefaultBranchProtectedKPI, 0.5)
        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.5)
        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)

        val rootKPI = KPICreateDto(
            "Project Score",
            "Assesses the project resp. the provided software in the aspects of maturity (based on quality, security and usability aspects) as well as development process.",
            true,
            AggregationKPICalculationStrategy()
        )
        rootKPI.addChildKPI(processTransparencyKPI, 0.5)
        rootKPI.addChildKPI(processComplianceKPI, 0.5)
        return rootKPI
    }
}
