package de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort

import de.fraunhofer.iem.dataprovider.dependency.dto.DependencyCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityScoreCreateDto
import de.fraunhofer.iem.dataprovider.dependency.enumeration.VulnerabilityScoringSystemEnum
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ToolProcessTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.dto.OrtResultDto
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.AdvisorResultJson
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.OrtJson
import kotlinx.serialization.json.*
import java.io.IOException
import java.nio.file.Path
import java.util.*

class OrtTask(
    override val responseChannel: suspend (event: Event) -> Unit,
    override val resultPath: Path,
    override val repoId: UUID,
    private val dependencyService: DependencyService
) : ToolProcessTask<List<DependencyCreateDto>>() {
    override val flags: Array<String>
        get() = TODO("Not yet implemented")

    override suspend fun parseProcessResults(resultPath: Path): List<DependencyCreateDto> {
        val dependencies = mutableListOf<DependencyCreateDto>()
        val ortJson = getOrtResultsFromFile(resultPath)

        ortJson.packages.forEach { p ->
            if (p.nameAndVersion != null) {
                dependencies.add(DependencyCreateDto(name = p.nameAndVersion))
            }
        }

        ortJson.advisorResults.forEach { advisorResult ->

            var dependency = dependencies.find { it.name == advisorResult.identifier }

            if (dependency == null) {
                dependency = DependencyCreateDto(advisorResult.identifier)
                dependencies.add(dependency)
            }
            advisorResult.vulnerabilities?.forEach { vulnerability ->
                if (vulnerability.id != null) {


                    val vulnerabilityScoreDtos: List<VulnerabilityScoreCreateDto> =
                        vulnerability.references.mapNotNull { ref ->
                            if (ref.severity != null && ref.scoringSystem != null) {
                                val scoringSystem = VulnerabilityScoringSystemEnum.fromString(ref.scoringSystem)
                                scoringSystem?.let {
                                    return@mapNotNull VulnerabilityScoreCreateDto(severity = ref.severity, it)
                                }
                            }
                            return@mapNotNull null
                        }
                    val vulnerabilityDto = VulnerabilityCreateDto(vulnerability.id, vulnerabilityScoreDtos)
                    dependency.vulnerabilities.add(vulnerabilityDto)
                }
            }
        }

        return dependencies
    }

    override suspend fun storeResultsInDb(result: List<DependencyCreateDto>) {
        result.forEach { dependency ->
            dependencyService.save(dependency.toDbObject())
        }
    }

}

fun getHighestScore(results: List<AdvisorResultJson>): Double {
    var highestScore = 0.0
    results.forEach { res ->
        res.vulnerabilities?.forEach { vul ->
            vul.references?.forEach { ref ->
                if (ref != null) {
                    if (ref.scoringSystem?.contains("cvss") == true) {
                        val score = ref.severity?.toDouble() ?: 0.0
                        if (highestScore < score) {
                            highestScore = score
                        }
                    }
                }
            }
        }
    }
    return highestScore
}

fun getOrtResultsFromFile(resultPath: Path): OrtResultDto {
    val resFile = resultPath.toFile()

    if (resFile.exists()) {
        val results = mutableListOf<AdvisorResultJson>()
        val resString = resFile.readText()
        val json = Json { ignoreUnknownKeys = true }

        val rawJson = json.decodeFromString<OrtJson>(resString)
        rawJson.advisor?.results?.advisorResults?.forEach { (key, value) ->
            if (value is JsonArray) {
                value.forEach { res ->
                    if (res is JsonObject) {

                        val flat = buildJsonObject {
                            res.entries.forEach { (eKey, eValue) ->
                                put(eKey, eValue)
                            }
                            put("identifier", key)
                        }

                        val advisorResult = json.decodeFromJsonElement<AdvisorResultJson>(flat)
                        results.add(advisorResult)
                    }
                }
            }
        }
        val packages = rawJson.analyzer?.result?.packages?.toList() ?: listOf()
        return OrtResultDto(advisorResults = results, packages = packages)
    }

    throw IOException("File not found / Path is no file.")
}
