package de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.qodana.json

import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.sarif.json.*
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.decodeFromString
import kotlinx.serialization.json.Json
import java.io.IOException
import java.nio.file.Path

@Serializable
data class QodanaSarif(
    @SerialName("runs")
    val runs: List<Run>?
)

fun getQodanaSarifFromFilePath(resultPath: Path): SarifJson {
    val resFile = resultPath.toFile()
    if (resFile.exists()) {
        val resString = resFile.readText()
        val json = Json { ignoreUnknownKeys = true }
        return json.decodeFromString<QodanaSarif>(resString).asSarif()
    }

    throw IOException("File not found / Path is no file.")
}

fun Extension.asSarif(): List<SarifRuleJson> {
    return this.rules?.mapNotNull { rule ->
        if (rule.id != null) {
            SarifRuleJson(
                id = rule.id,
                shortDescription = SarifShortDescriptionJson(rule.fullDescription?.text ?: ""),
                name = rule.shortDescription?.text ?: ""
            )
        } else {
            null
        }
    } ?: listOf()
}

fun Tool.asSarif(): SarifToolJson {
    return SarifToolJson(
        SarifDriverJson(
            name = this.driver?.name,
            fullName = this.driver?.fullName,
            rules = this.extensions?.flatMap { it.asSarif() } ?: listOf(),
            version = this.driver?.version
        )
    )
}

fun Region.asSarif(): SarifRegionJson {
    return SarifRegionJson(startLine = this.startLine ?: -1, startColumn = this.startColumn ?: -1)
}

fun Location.asSarif(): SarifLocationJson {
    return SarifLocationJson(
        SarifPhysicalLocationJson(
            artifactLocation = SarifArtifactLocationJson(this.physicalLocation?.artifactLocation?.uri ?: ""),
            region = this.physicalLocation?.region?.asSarif()
        )
    )
}

fun QodanaPromoResult.asSarif(): SarifResultJson? {
    if (this.ruleId != null) {
        return SarifResultJson(
            ruleId = this.ruleId,
            level = this.level ?: "",
            locations = this.locations?.map { it.asSarif() } ?: listOf(),
            message = SarifMessageJson(this.message?.text ?: ""),
        )
    }
    return null
}

fun Run.asSarif(): SarifRunJson? {
    val tool = this.tool?.asSarif()
    val results = this.properties?.qodanaPromoResults?.mapNotNull { result -> result.asSarif() }
        ?: listOf()
    if (tool != null) {
        return SarifRunJson(tool = tool, results = results)
    }

    return null
}

fun QodanaSarif.asSarif(): SarifJson {
    val sarifRuns = this.runs?.mapNotNull { it.asSarif() }
    return SarifJson(sarifRuns ?: listOf())
}